from __future__ import annotations
import warnings
from typing_extensions import SupportsBytes, SupportsIndex
from types import GenericAlias
from typing import (
    Generic,
    Iterable,
    TypeAlias,
    TypeVar,
    Self,
    Literal,
    LiteralString,
    cast,
    get_args,
    Union,
    get_origin,
    Any,
    overload,
)
from pydantic import ValidatorFunctionWrapHandler
from pydantic_core import PydanticSerializationUnexpectedValue
from pydantic_core import core_schema as cs
from pydantic.annotated_handlers import GetCoreSchemaHandler

from ._guard import require, verify, NotEqualError

from ._typeshed import ReadableBuffer
from ._numeric import negative, positive
from ._hidden import LiteralGenericAlias


Nothing = Literal["~"]
NOTHING: Nothing = get_args(Nothing)[0]
# TODO: move to literal module, ensure conformance with mathematical standards
# MathematicalOperation = Literal[
# "addition",
# "subtraction",
# "multiplication",
# "division",
# "exponentation",
# ... etc
# ]
# FIXME: incomplete
MathematicalOperation = Literal["+", "-", "*", "/", "**"]
ComparisonOperation = Literal["==", "!=", ">=", ">", "<=", "<"]
Operation = (
    MathematicalOperation | ComparisonOperation
)  # repl str with comp op

# T = TypeVar("T", bound=LiteralString | Nothing)
T = TypeVar("T", bound=LiteralString)


class DifferentThingsError(NotEqualError, TypeError):
    """Raised when a mathematical operation is run on two different things."""

    def __init__(
        self,
        source: Any,
        target: Any,
        detail: str | None,
        operation: Operation,
        *args: object,
    ) -> None:
        self.operation = operation
        detail = f"The operation '{operation}' cannot be used on counts of different things."
        super().__init__(source, target, detail, *args)


# class AlreadyAThingError(Exception): ...


# class arguments:
#     """Namespace for internal argument normalization functions."""

#     @staticmethod
#     def norm_tuple(): ...


class toolkit:
    @staticmethod
    def validate_same_thing(a: count, b: count, operation: Operation) -> None:
        require.eq(
            source=a.__thing__,
            target=b.__thing__,
            raises=DifferentThingsError,
            operation=operation,
        )


# FIXME: using whole seems to be breaking pydantic atm
# class count(whole, Generic[T]):
class count(int, Generic[T]):
    """
    A count `n` of of a `thing` (widgets, bytes, cities, etc).
    Used in contexts where you want to count a thing and have
    the count for that thing be differentiable from another
    thing being counted. Only for whole numbers, no decimals/floats.

    ### Overview

    ```python
    ```

    ### Details

    What is the purpose? Take this example:

    ```python
    import pydantic

    class LLM(pydantic.BaseModel):
        size_gigabytes: int
        num_parameters: int
        quantization_bits: int

    llm = LLM(size_gigabytes=10, num_parameters=10, quantization_bits=10)

    assert llm.size_gigabytes == llm.num_parameters, "This will pass but shouldn't."
    ```

    The assertion above shouldn't pass because gigabytes are not parameters,
    plus, the information about the fact that the size in gigabytes and
    quantization in bits is only captured in the variable/attribute name.

    The `count` type solves the first issue and partially solves the second
    issue as follows:

    ```python
    import pydantic
    from typing import Literal

    class LLM(pydantic.BaseModel):
        size: count[Literal["gigabyte"]]
        parameters: count[Literal["parameter"]]
        quantization: count[Literal["bit"]]

    llm = LLM(
        size=count(10, "gigabyte"),
        parameters=count((10, "parameter")),
        quantization=count(value=10, thing="bit")),
    )

    assert llm.size != llm.parameters, "This will pass, and should."

    # Note that if you do something like:

    llm = LLM(
        size=count(10, "byte"),
        parameters=count((10, "params")),
        quantization=count(n=10, thing="bits")),
    )

    # your static type checker should catch it, and
    # additionally pydantic will raise an exception
    ```

    Additionally, if you get tired of passing in literals,
    it's straightforward to define your own count types of
    arbitrary things:

    ```python
    TODO
    ```

    If you want to do more detailed unit/quantity
    modeling and conversion, check out: [pint](https://github.com/hgrecco/pint)

    """

    __defined__: bool = False
    # __thing__: T | Nothing = "~"
    __thing__: T

    # :common
    def __str__(self) -> str:
        return f"{super().__str__()} [{self.thing}]"

    def __repr__(self) -> str:
        return f"{super().__repr__()} [{self.thing}]"

    # :constructor
    # ! if the __thing__ is already not none, then passing a new thing should raise an error

    @overload
    def __new__(cls, n: int) -> count[T]:
        """_summary_

        Args:
            n (int): _description_

        Returns:
            count[Nothing]: _description_
        """
        ...

    @overload
    def __new__(cls, n: int, thing: T) -> count[T]:
        """_summary_

        Args:
            n (int): _description_
            thing (T): _description_

        Returns:
            count[T]: _description_
        """
        ...

    @overload
    def __new__(cls, value: ConvertibleToCount) -> count[T]:
        """_summary_

        Args:
            value (ConvertibleToCount): _description_

        Returns:
            count[T]: _description_
        """
        ...

    def __new__(cls, *args, **kwargs) -> count[T]:

        match (len(args), len(kwargs)):
            case (1, 0):
                n = verify.isinstance(args[0], int)
                thing = NOTHING
            case (2, 0):  # count(10, "bit")
                n = verify.isinstance(args[0], int)
                thing = verify.isinstance(args[1], str)
                require.ne(
                    thing,
                    NOTHING,
                    f"'{NOTHING}' is reserved for the absence of a thing and cannot be used here.",
                )
            case _:
                raise NotImplementedError("TODO")
        require.ge(n, 0)
        inst = int.__new__(cls, n)
        inst.__thing__ = cast(T, thing)
        return inst

    @property
    def n(self) -> int:
        """_summary_

        Returns:
            int: _description_
        """
        return int(self)

    @property
    def thing(self) -> T:
        """_summary_

        Returns:
            T: _description_
        """
        return cast(T, self.__thing__)

    @property
    def value(self) -> tuple[int, T]:
        """_summary_

        Returns:
            tuple[int, T]: _description_
        """
        return (self.n, self.thing)

    @classmethod
    def __serialize__(
        cls,
        value: count[T],
        info: cs.SerializationInfo,
    ) -> tuple[int, T] | Self:
        if not isinstance(value, cls):
            raise PydanticSerializationUnexpectedValue(
                f"Expected `{cls}` but got `{type(value)}` with value "
                f"`'{value}'` - serialized value may not be as expected."
            )
        return cast(tuple[int, T], tuple(value))

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: type[count[T]],
        handler: GetCoreSchemaHandler,
    ) -> cs.CoreSchema:
        def wrapper(
            # TODO: add int - should only be coerced if source is defined count
            value: ConvertibleToCount,
            wrap_handler: ValidatorFunctionWrapHandler,
        ):
            # * source: count | value: count(10, "gram")
            if source is count:
                return count(value)
                # return source.__new__(source, value)
            else:
                internal = count(value)
                if isinstance(
                    source, GenericAlias
                ):  # > count[Literal["gram"]]
                    if get_origin(source) is count:
                        generic_args = get_args(source)  # >(Literal['gram'],)
                        if len(generic_args) != 1:
                            raise TypeError(source)
                        literal_type = generic_args[0]  # > Literal['gram']
                        if not isinstance(literal_type, LiteralGenericAlias):
                            raise TypeError(source)
                        literal_args = get_args(literal_type)  # > ('gram',)
                        if len(literal_args) != 1:
                            raise TypeError(source)
                        if not isinstance(literal_args[0], str):
                            raise TypeError(source)
                        source_thing = literal_args[0]
                        if internal.thing != source_thing:
                            raise TypeError(
                                f"'{internal.thing}' must be '{source_thing}'"
                            )
                        return count(internal.n, cast(T, source_thing))
                    else:
                        raise TypeError(source)
                elif (source.__base__ is count) and (
                    getattr(source, "__defined__")
                ):
                    require.eq(source.__thing__, internal.__thing__)  # type: ignore
                    return source.__new__(source, internal.n)
                else:
                    if (value.__class__.__base__ is count) and (
                        getattr(value.__class__, "__defined__")
                    ):
                        # * handle custom unit
                        # return count(value_n, value_thing)
                        pass
                    else:
                        pass
                        # return count(value_n, value_thing)
                raise NotImplementedError()

        return cs.no_info_wrap_validator_function(
            wrapper,
            schema=cs.tuple_schema(
                [
                    cs.int_schema(
                        ge=0,
                        strict=True,
                    ),
                    cs.str_schema(
                        min_length=0,
                        max_length=1024,
                        strict=True,
                    ),
                ],
                # variadic_item_index=1,
                min_length=2,
                max_length=2,
                strict=True,
            ),
            serialization=cs.plain_serializer_function_ser_schema(
                cls.__serialize__, info_arg=True, when_used="always"
            ),
        )

    @staticmethod
    # def define(thing: T) -> type[count[T]]:
    def define(thing: T):
        """
        Define a new type for counting something.
        """
        # TODO: move main code to sep func

        class count_of_thing(count, Generic[T]):  # type: ignore
            """
            A custom defined count of a thing.
            """

            __defined__ = True
            __thing__: T = thing

            # def __new__(cls, n: int) -> count[T]:
            def __new__(cls, n: int) -> count_of_thing:
                return super().__new__(cls, n, thing)  # type: ignore

        return type("count_of_thing", (count_of_thing,), {})
        # return type(f"count_of_{thing}", (count_of_thing,), {})

    # :operation

    def __add__(self, value: int | count[T], /) -> count[T]:
        if isinstance(value, count):
            toolkit.validate_same_thing(self, value, "+")
            nsum = super().__add__(value.n)
            inst = self.__class__(nsum)
            inst.__thing__ = self.thing
            return inst
        elif isinstance(value, int):
            warnings.warn("count:int operations may be deprecated")
            nsum = super().__add__(value)
            inst = self.__class__(nsum)
            inst.__thing__ = self.thing
            return inst
        else:
            raise TypeError(value)

    def __radd__(self, value: int | count[T], /) -> count[T]:
        return self.__add__(value)

    def __iadd__(self, value: int | count[T], /) -> count[T]:
        return self.__add__(value)

    # TODO: clean up redundant code for unary/binary ops
    def __sub__(self, value: int | count[T], /) -> count[T]:
        if isinstance(value, count):
            toolkit.validate_same_thing(self, value, "-")
            nsum = super().__sub__(value.n)
            inst = self.__class__(nsum)
            inst.__thing__ = self.thing
            return inst
        elif isinstance(value, int):
            warnings.warn("count:int operations may be deprecated")
            nsum = super().__sub__(value)
            inst = self.__class__(nsum)
            inst.__thing__ = self.thing
            return inst
        else:
            raise TypeError(value)

    def __rsub__(self, value: int | count[T], /) -> count[T]:
        return self.__sub__(value)

    def __isub__(self, value: int | count[T], /) -> count[T]:
        return self.__sub__(value)

    def __mul__(self, value: int, /) -> count[T]:  # type: ignore
        """
        *Note that the whole number enforcement is currently broken.*
        Counts are whole numbers, and can only be multiplied by whole numbers.
        Technically, if you have something like:

        ```python
        a = count(10, "meter")
        b = count(10, "meter")
        a * b
        ```

        That would equal something like 'meters squared', but this type
        doesn't do unit conversion, and is for counting arbitrary things
        (meters, widgets, thoughts, colors, notes, caffeine pills, etc);
        because `10 [thoughts] * 10 [thoughts] != 100 [thoughts^2]` (this
        is nonsensical), multiplication only works
        """
        require.isinstance(value, int)
        nsum = super().__mul__(value)
        inst = self.__class__(nsum)
        inst.__thing__ = self.thing
        return inst

    def __rmul__(self, value: int, /) -> count[T]:  # type: ignore
        return self.__mul__(value)

    def __imul__(self, value: int, /) -> count[T]:
        return self.__mul__(value)

    # NOTE: should maybe use __or__ and explicitly return ratio type
    def __truediv__(self, value: count[T], /) -> float:  # type: ignore
        """
        Counts can only be divided by other counts of the same thing,
        which cancels out the count and returns a float that represents
        the ratio between the counts.

        This decision was made so that count values are enforced as whole numbers;
        division is likely to result in non-integer numbers, so while
        technically `10 [gram] / 5 == 2 [gram]` and `10 [g] / 3 == 3.33 [g]`,
        allowing integers would also add cases like `3 [colors] / 2 == 1.5 [colors]`;
        if you had `red, green, blue` and you divided it by `red, green`,
        would this result in `red, gre`? Maybe `red, yellow`? Doesn't make sense.

        """
        require.isinstance(value, count)
        toolkit.validate_same_thing(self, value, "/")
        return self.n / value.n

    def __eq__(self, value: object, /) -> bool:
        if not isinstance(value, count):
            return False
        return (self.thing == value.thing) and (self.n == value.n)

    def __ne__(self, value: object, /) -> bool:
        return not self.__eq__(value)

    # TODO: can consolidate this code with operator, repetitive
    def __lt__(self, value: count[T], /) -> bool:  # type: ignore
        toolkit.validate_same_thing(self, value, "<")
        return self.n < value.n

    def __le__(self, value: count[T], /) -> bool:  # type: ignore
        """"""
        toolkit.validate_same_thing(self, value, "<=")
        return self.n <= value.n

    def __gt__(self, value: count[T], /) -> bool:  # type: ignore
        """"""
        toolkit.validate_same_thing(self, value, ">")
        return self.n > value.n

    def __ge__(self, value: count[T], /) -> bool:  # type: ignore
        """"""
        toolkit.validate_same_thing(self, value, ">=")
        return self.n >= value.n

    def __float__(self) -> float:
        return super().__float__()

    def __int__(self) -> int:
        return super().__int__()

    def __hash__(self) -> int:
        return hash((self.n, self.thing))

    def __bool__(self) -> bool:
        warnings.warn("unstable")  # repl with decorator
        return self.n != 0

    def __iter__(self):
        for x in (self.n, self.thing):
            yield x

    # TODO.backlog

    def __abs__(self) -> int: ...
    def as_integer_ratio(self) -> tuple[int, Literal[1]]: ...
    @property
    def real(self) -> int: ...
    @property
    def imag(self) -> Literal[0]: ...
    @property
    def numerator(self) -> int: ...
    @property
    def denominator(self) -> Literal[1]: ...
    def conjugate(self) -> int: ...
    def bit_length(self) -> int: ...
    def bit_count(self) -> int: ...
    def to_bytes(
        self,
        length: SupportsIndex = 1,
        byteorder: Literal["little", "big"] = "big",
        *,
        signed: bool = False,
    ) -> bytes: ...
    @classmethod
    def from_bytes(
        cls,
        bytes: Iterable[SupportsIndex] | SupportsBytes | ReadableBuffer,
        byteorder: Literal["little", "big"] = "big",
        *,
        signed: bool = False,
    ) -> Self: ...

    def is_integer(self) -> Literal[True]:
        raise NotImplementedError(
            "subclass of integer, but not technically integer - undetermined"
        )

    # def is_whole(self):
    #     raise NotImplementedError("same issue as with integer")

    def is_count(self) -> Literal[True]:
        return True

    def __floordiv__(self, value: int, /) -> int: ...
    def __mod__(self, value: int, /) -> int: ...
    def __divmod__(self, value: int, /) -> tuple[int, int]: ...
    def __rfloordiv__(self, value: int, /) -> int: ...
    def __rtruediv__(self, value: int, /) -> float: ...
    def __rmod__(self, value: int, /) -> int: ...
    def __rdivmod__(self, value: int, /) -> tuple[int, int]: ...
    @overload
    def __pow__(self, x: Literal[0], /) -> Literal[1]: ...  # type: ignore
    @overload
    def __pow__(self, value: Literal[0], mod: None, /) -> Literal[1]: ...
    @overload
    def __pow__(self, value: positive[int], mod: None = None, /) -> int: ...
    @overload
    def __pow__(self, value: negative[int], mod: None = None, /) -> float: ...
    @overload
    def __pow__(self, value: int, mod: None = None, /) -> Any: ...
    @overload
    def __pow__(self, value: int, mod: int, /) -> int: ...  # type: ignore
    def __rpow__(self, value: int, mod: int | None = None, /) -> Any: ...
    def __and__(self, value: int, /) -> int: ...
    def __or__(self, value: int, /) -> int: ...
    def __xor__(self, value: int, /) -> int: ...
    def __lshift__(self, value: int, /) -> int: ...
    def __rshift__(self, value: int, /) -> int: ...
    def __rand__(self, value: int, /) -> int: ...
    def __ror__(self, value: int, /) -> int: ...
    def __rxor__(self, value: int, /) -> int: ...
    def __rlshift__(self, value: int, /) -> int: ...
    def __rrshift__(self, value: int, /) -> int: ...
    def __neg__(self) -> int: ...
    def __pos__(self) -> int: ...
    def __invert__(self) -> int: ...
    def __trunc__(self) -> int: ...
    def __ceil__(self) -> int: ...
    def __floor__(self) -> int: ...
    def __round__(self, ndigits: SupportsIndex = ..., /) -> int: ...
    def __getnewargs__(self) -> tuple[int]: ...
    def __index__(self) -> int: ...


ConvertibleToCount = Union[
    # int,
    # str,
    count[T],
    # tuple[int],
    tuple[int, T],
    # list[int | T],
]
"""Union of all types parsable to `count`."""

# from typing import Literal as L

# ByteCount: type[count[L["byte"]]] = count[L["byte"]].define("byte")
# BitCount = count[L["bit"]].define("bit")
# FooCount: type[count[L["foo"]]] = count.define("foo")

# def f():
#     return BitCount(10)

# def pylance():
#     # test with literal type alias
#     # gram = count[Literal["gram"]]

#     from pydantic import BaseModel

#     class Test(BaseModel):
#         x: BitCount





"""
Useful types, not sure if they're 'hidden', but pylance struggles to recognize them.

If this module causes errors then LMK on github cause I think they would be fixed
with some try/except blocks, but don't want to both right now. EG:

```python
try:
    LiteralGenericAlias = typing._LiteralGenericAlias  # type: ignore
except:
    LiteralGenericAlias = type(typing.Literal[""])
```
"""
import typing

LiteralGenericAlias = typing._LiteralGenericAlias  # type: ignore
Sentinel = typing._Sentinel # type: ignore
Final = typing._Final # type: ignore
NotIterable = typing._NotIterable # type: ignore
SpecialForm = typing._SpecialForm # type: ignore
TypedCacheSpecialForm = typing._TypedCacheSpecialForm # type: ignore
AnyMeta = typing._AnyMeta # type: ignore
BaseGenericAlias = typing._BaseGenericAlias # type: ignore
GenericAlias = typing._GenericAlias # type: ignore
SpecialGenericAlias = typing._SpecialGenericAlias # type: ignore

# TODO: add others as needed


from __future__ import annotations
import builtins
import orjson
from abc import ABC, abstractmethod
from datetime import datetime
from types import ModuleType
from typing import Any, TypeVar, overload, ClassVar
from typing import Generic, Iterable, TypeVar, Self
from pydantic import BaseModel, ConfigDict, Json, Field
from pydantic_core import core_schema, PydanticSerializationUnexpectedValue
from pydantic.annotated_handlers import (
    GetJsonSchemaHandler,
    GetCoreSchemaHandler,
)
from pydantic.json_schema import JsonSchemaValue


Model = BaseModel


class Documented(Model):
    model_config = ConfigDict(use_attribute_docstrings=True)


class Blob(Model): ...



from __future__ import annotations
from typing import Iterable, SupportsIndex, overload
from typing_extensions import LiteralString
from typically import literal


class estr(str):

    # @property
    # def encoding(self) -> literal.CharacterEncoding: ...

    def isuuid(self, version: literal.UUIDVersion) -> bool: ...

    # def tocase(self, casing: ...): ...

    def dedent(self) -> estr: ...
    def escape(self) -> estr: ...
    def apply(self): ...

    # TODO: regex utils

    @overload
    def replace(
        self: LiteralString,
        old: LiteralString,
        new: LiteralString,
        /,
        count: SupportsIndex = -1,
    ) -> LiteralString: ...
    @overload
    def replace(
        self,
        old: str,
        new: str,
        /,
        count: SupportsIndex = -1,
    ) -> str: ...  # type: ignore[misc]
    @overload
    def replace(self, mapping: dict[str, str]) -> str: ...  # TODO
    def replace(self, *args, **kwargs) -> str:  # type: ignore
        return super().replace(*args, **kwargs)

    def fuzzysearch(
        self, query: str, window: int | None = None, n: int = 1
    ) -> _fuzzy_search_result | list[_fuzzy_search_result]: ...

    @overload
    def cjoin(
        self: LiteralString, iterable: Iterable[LiteralString], /
    ) -> estr: ...
    @overload
    def cjoin(self, iterable: Iterable[str], /) -> estr: ...  # type: ignore[misc]
    def cjoin(
        self: LiteralString | str | estr,
        iterable: Iterable[str] | Iterable[LiteralString],
        /,
        exclude_none: bool = True,
    ) -> estr: ...


# class _estr_slice(str): ...
class _fuzzy_search_result(estr): ...



from __future__ import annotations
import rpi.v0 as rpi
import inspect
import importlib
from pathlib import Path
from types import ModuleType
from typing import Union, Callable, Any, Self
from pydantic import ValidatorFunctionWrapHandler
from pydantic.annotated_handlers import GetCoreSchemaHandler
from pydantic_core import PydanticSerializationUnexpectedValue
from pydantic_core import core_schema as cs

# TODO: add classvar and model type for web API inputs/payloads that together allow blocking
# of some types for use in apis - eg: definition
# TODO: cover failure cases of inspect.getsource for C/pyo3 modules, or dynamic runtime generated code


class module:
    @staticmethod
    def __validate_str(value: str) -> str:
        value = value.strip()
        if rpi.name_chain.fullmatch(value) is None:
            raise ValueError(
                f"'{value}' does not match pattern r'{rpi.name_chain.pattern}'"
            )
        return value

    @staticmethod
    def __validate_mod(value: ModuleType) -> str:
        return value.__name__

    @staticmethod
    def __validate_value(value: str | ModuleType) -> str:
        if isinstance(value, str):
            return module.__validate_str(value)
        elif isinstance(value, ModuleType):
            return module.__validate_mod(value)
        else:
            raise TypeError(value)

    @staticmethod
    def __init_object(value: str | ModuleType) -> ModuleType | None:
        return None if isinstance(value, str) else value

    def __init__(self, value: str | ModuleType) -> None:
        self.__value = self.__validate_value(value)
        self.__object = self.__init_object(value)

    def __str__(self) -> str:
        return self.__value

    def __repr__(self) -> str:
        return self.__str__()

    def reset(self) -> None:
        """Reset cached data."""
        self.__object = None

    @property
    def name(self) -> str:
        return str(self)

    # TODO: review security implications of dynamic import
    @property
    def object(self) -> ModuleType:
        """
        The module object itself.
        """
        if self.__object is None:
            self.__object = importlib.import_module(name=self.name)
        return self.__object

    @property
    def doc(self) -> str:
        """
        The module object docstring.
        """
        return self.object.__doc__ or ""

    @property
    def exists(self) -> bool:
        """
        Whether the module is available in the current environment.
        """
        # return self.value in sys.modules # won't work for relative?
        try:
            _ = self.object
            return True
        except:
            return False

    @property
    def code(self) -> str:
        return inspect.getsource(self.object)

    @property
    def file(self) -> Path | None:
        __file__ = self.object.__file__
        if __file__ is None:
            return None
        # path = Path(__file__).absolute()
        path = Path(__file__).resolve()
        return path

    @property
    def path(self) -> Path | None:
        __path__ = self.object.__path__
        if len(__path__) == 1:
            path = Path(__path__[0]).resolve()
            return path
        else:
            return None

    @staticmethod
    def factory(value: ConvertibleToModule) -> Callable[[], module]:
        return lambda: module(value)

    @classmethod
    def __serialize__(  # type: ignore
        cls,
        value: module,
        info: cs.SerializationInfo,
    ) -> str | Self:
        if not isinstance(value, cls):
            raise PydanticSerializationUnexpectedValue(
                f"Expected `{cls}` but got `{type(value)}` with value `'{value}'` - serialized value may not be as expected."
            )
        if info.mode == "json":
            return str(value)
        return value

    @classmethod
    def __get_pydantic_core_schema__(  # type: ignore
        cls,
        source: type[module],
        handler: GetCoreSchemaHandler,
    ) -> cs.CoreSchema:
        def wrapper(value: Any, wrap_handler: ValidatorFunctionWrapHandler):
            return source(value)

        return cs.no_info_wrap_validator_function(
            wrapper,
            schema=cs.str_schema(
                pattern=rpi.name_chain,
                min_length=1,
                max_length=128,
                strip_whitespace=True,
                strict=True,
            ),
            serialization=cs.plain_serializer_function_ser_schema(
                cls.__serialize__, info_arg=True, when_used="always"
            ),
        )


class definition:
    @staticmethod
    def __validate_str(value: str) -> tuple[str, str]:
        m = rpi.definition_locator.fullmatch(value)
        if m is None:
            raise ValueError(
                f"'{value}' is not in form of '<module>:<object>', "
                f"does not match pattern r'{rpi.definition_locator.pattern}'"
            )
        return (m.group(1), m.group(4))

    @staticmethod
    def __validate_def(value: definition) -> tuple[str, str]:
        return definition.__validate_str(str(value))

    @staticmethod
    def __validate_tuple(value: tuple[str, str]) -> tuple[str, str]:
        if len(value) != 2:
            raise ValueError(value)
        if not all(isinstance(item, str) for item in value):
            raise TypeError(value)
        return definition.__validate_str(":".join(value))

    @staticmethod
    def __validate_value(value: ConvertibleToDefinition) -> tuple[str, str]:
        if isinstance(value, ModuleType):
            raise TypeError(
                f"'{value.__name__}' is a ModuleType, "
                f"'{module.__module__}:{module.__qualname__}' should be used instead"
            )
        else:
            if isinstance(value, str):
                return definition.__validate_str(value)
            elif isinstance(value, definition):
                return definition.__validate_def(value)
            elif isinstance(value, tuple):
                return definition.__validate_tuple(value)
            # FIXME: inspect.isfunction is not reliable, should be replaced
            elif inspect.isfunction(value) or isinstance(value, type):
                return (value.__module__, value.__qualname__)
            else:
                raise TypeError(value)

    @staticmethod
    def __init_object(value: ConvertibleToDefinition) -> object | None:
        return None if isinstance(value, str | tuple | definition) else value

    def __init__(self, value: ConvertibleToDefinition) -> None:
        self.__value = self.__validate_value(value)
        self.__object = self.__init_object(value)

    def __str__(self) -> str:
        return self.locator

    def __bool__(self) -> bool:
        return self.exists

    def __repr__(self) -> str:
        return self.locator

    def reset(self) -> None:
        """Reset cached object."""
        self.__object = None

    @property
    def value(self) -> tuple[str, str]:
        return self.__value

    @property
    def module(self) -> module:
        return module(self.value[0])

    @property
    def exists(self) -> bool:
        """
        Check if the definition exists.
        """
        try:
            _ = self.object
            return True
        except AttributeError:
            return False

    @property
    def name(self) -> str:
        """
        Object qualified name in module.

        Examples:
            - `"urlparse"`
            - `"geohash"`
            - `"definition.locationlocator`
        """
        return self.value[1]

    @property
    def doc(self) -> str | None:
        """
        The definition object docstring.
        """
        return self.object.__doc__

    @property
    def code(self) -> str:
        """Get source code for object."""
        return inspect.getsource(self.object)  # type: ignore

    @property
    def object(self) -> object:
        """
        Get defined object.
        """
        if self.__object is None:
            target = self.module.object
            scopes: list[str] = self.name.split(".")
            while len(scopes):
                name = scopes.pop(0)
                target = getattr(target, name)
            self.__object = target
        return self.__object

    @property
    def type(self) -> type:
        """
        Get object type. Probably will be `function` or `type` (`type` is an instance of itself).
        """
        return type(self.object)

    @property
    def locator(self) -> str:  # ? identity, location, source
        """
        Locator/location: joined name in form of: `<module_name>:<definition_name>`.

        Examples:
            - `"urllib.parse:urlparse"`
            - `"antigravity:geohash"`
            - `"typically.types:definition.locator"` (this property)
        """
        return ":".join(self.value)

    @property
    def cls(self) -> bool:
        """Whether the object is a class."""
        return inspect.isclass(self.object)

    @property
    def fn(self) -> bool:
        """Whether the object is a function."""
        # TODO: replace with reliable inspection func
        return inspect.isfunction(self.object)

    # TODO: add other high level type checks (is instance, is variable?, etc)

    @staticmethod
    def factory(value: ConvertibleToDefinition) -> Callable[[], definition]:
        return lambda: definition(value)

    @classmethod
    def __serialize__(  # type: ignore
        cls,
        value: definition,
        info: cs.SerializationInfo,
    ) -> str | definition:
        if not isinstance(value, cls):
            raise PydanticSerializationUnexpectedValue(
                f"Expected `{cls}` but got `{type(value)}` with value `'{value}'` - serialized value may not be as expected."
            )
        if info.mode == "json":
            return str(value)
        return definition(value)

    @classmethod
    def __get_pydantic_core_schema__(  # type: ignore
        cls,
        source: type[definition],
        handler: GetCoreSchemaHandler,
    ) -> cs.CoreSchema:
        def wrapper(value: Any, wrap_handler: ValidatorFunctionWrapHandler):
            return definition(value)

        return cs.no_info_wrap_validator_function(
            wrapper,
            schema=cs.str_schema(
                pattern=rpi.definition_locator,
                min_length=1,
                max_length=256,
                strip_whitespace=True,
                strict=True,
            ),
            serialization=cs.plain_serializer_function_ser_schema(
                cls.__serialize__,
                info_arg=True,
                when_used="always",
            ),
        )


ConvertibleToDefinition = Union[
    str,
    tuple[str, str],
    definition,
    object,
]
ConvertibleToModule = Union[
    str,
    ModuleType,
]



from __future__ import annotations
import re
import inspect
import importlib
from decimal import Decimal, DecimalTuple, Context
from types import ModuleType as module
from typing import (
    Any,
    Generic,
    TypeVar,
    overload,
    get_args,
    TypeVar,
    Self,
    ClassVar,
    cast,
    Callable,
)
from typing_extensions import LiteralString
from pydantic import ValidatorFunctionWrapHandler
from pydantic_core import core_schema, PydanticSerializationUnexpectedValue
from pydantic.annotated_handlers import GetCoreSchemaHandler

# TODO: split into submodules once type interdependency graph is clearer
# TODO: may be useful to split numeric into scalar/vector modules
# TODO: reimplement fractions.Fraction + reciprocal prop

T = TypeVar("T")
P = TypeVar("P")
CT = TypeVar("CT", covariant=True)


ordered = list[T] | tuple[T, ...]


class pair(tuple[T, T]):
    @classmethod
    def __new_args__(cls, *args) -> pair[T]:
        match len(args):
            case 1:
                arg = args[0]
                if len(arg) == 2:
                    return tuple.__new__(cls, arg)
                else:
                    raise TypeError(
                        f"'{cls.__name__}' expects 2 args or an iterable of length 2, but received length {len(arg)}: '{arg}'",
                    )
            case 2:
                return tuple.__new__(cls, args)
            case _:
                raise TypeError(args)

    @classmethod
    def __new_kwargs__(cls, **kwargs) -> pair[T]:
        # FIXME
        keys = set(kwargs.keys())
        if keys == {"value"}:
            return cls.__new_args__(*kwargs["value"])
        elif keys == {"a", "b"}:
            return cls.__new_args__(kwargs["a"], kwargs["b"])
        else:
            raise TypeError(f"Invalid keyword arguments: {kwargs}")

    @overload
    def __new__(cls, value: pair[T] | tuple[T, T] | list[T]) -> pair[T]: ...
    @overload
    def __new__(cls, a: T, b: T) -> pair[T]: ...
    def __new__(cls, *args, **kwargs) -> pair[T]:
        if len(args):
            return cls.__new_args__(*args)
        elif len(kwargs):
            return cls.__new_kwargs__(**kwargs)
        else:
            raise TypeError(
                f"{cls.__name__} expects a pair of two items, or an iterable of length 2",
                args,
                kwargs,
            )

    def __hash__(self) -> int:
        return hash((self.a, self.b))

    @property
    def a(self) -> T:
        return self[0]

    @property
    def b(self) -> T:
        return self[1]

    @classmethod
    def __serialize__(
        cls,
        value: pair[T],
        info: core_schema.SerializationInfo,
    ) -> tuple[T, ...] | Self:
        if not isinstance(value, cls):
            raise PydanticSerializationUnexpectedValue(
                f"Expected `{cls}` but got `{type(value)}` with value `'{value}'` - serialized value may not be as expected."
            )
        if info.mode == "json":
            return tuple(value)
        return value

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source: type[pair[T]],
        handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        def wrapper(value: Any, wrap_handler: ValidatorFunctionWrapHandler):
            generic_param_types: tuple[type, ...] = get_args(source)
            match len(generic_param_types):
                case 0:
                    return source.__new__(source, value)
                case 1:
                    generic_type: type = generic_param_types[0]
                    new_value = list()
                    for item in value:
                        if isinstance(item, generic_type):
                            new_value.append(item)
                        else:
                            try:
                                new_value.append(generic_type(item))
                            except ValueError:
                                raise TypeError(
                                    f"Could not coerce value '{item}' to type '{generic_type.__name__}'"
                                )
                    return source.__new__(source, value.__class__(new_value))
                case _:
                    raise TypeError(
                        f"'pair' can only have one generic type, but has: '{generic_param_types}'"
                    )

        return core_schema.no_info_wrap_validator_function(
            wrapper,
            schema=core_schema.tuple_schema(
                [core_schema.any_schema(), core_schema.any_schema()],
                variadic_item_index=1,
                min_length=2,
                max_length=2,
                strict=True,
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls.__serialize__, info_arg=True, when_used="always"
            ),
        )


# TODO: real needs pydantic schema impl.
# TODO: reimplement in rust
# TODO: add precision/precise number of significant figures
# TODO: optimize, lot of bloat
# ? real, precise, decimal
class decimal(Decimal):
    """
    Same as decimal but with greater interoperability across other numeric types.

    NOTE: `float` to `str` cast
        Instantiating `Decimal(3.14)` yields the decimal:

        `Decimal('3.140000000000000124344978758017532527446746826171875')`

        This is probably because even though `Decimal` has arbitrary precision,
        the float passed in doesn't. As such input floats are converted to strings
        before being passed to the constructor.

    NOTE: Infinity
        ```python
        x = Decimal(0) / Decimal('inf')
        ```

        In the code above, `x == Decimal('0E-1000026')`.
        Accordingly infinity has just been blocked from being a valid value.
        May re-evaluate later.
    """

    # TODO: test with numpy/sympy types, make sure this works w/ array, sparse matrices, etc

    # * core methods

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__str__()})"

    def __hash__(self) -> int:
        return super().__hash__()

    def __copy__(self) -> decimal:
        return decimal(super().__copy__())

    def __deepcopy__(self, memo: Any) -> decimal:
        return decimal(super().__deepcopy__(memo))

    # * constructors

    def __new__(
        cls,
        value: Realish,
        context: Context | None = None,
    ) -> decimal:
        if isinstance(value, float):
            value = str(value)
        lstr = str(value).lower()
        if "inf" in lstr:
            raise TypeError(f"'{cls.__name__}' does not allow infinite value")
        if "nan" in lstr:
            raise TypeError(f"'{cls.__name__}' does not allow nan values")
        return Decimal.__new__(cls, value=value, context=context)

    @classmethod
    def from_numeric(cls, value: Numeric) -> decimal:
        if isinstance(value, str):
            raise TypeError(f"{type(value)}: '{value}'")
        return cls(value)

    @classmethod
    def from_float(cls, f: float) -> Self:
        return cls(str(f))

    # * type conversion

    def __bool__(self) -> bool:
        return super().__bool__()

    def __float__(self) -> float:
        return super().__float__()

    def __int__(self) -> int:
        return super().__int__()

    def __complex__(self) -> complex:
        return super().__complex__()

    @property
    def decimal(self) -> Decimal:
        return Decimal(self)

    def as_integer_ratio(
        self,
    ) -> tuple[int, int]: ...  # TODO: replace with ratio / pair[int]

    # * formatting

    def to_eng_string(self, context: Context | None = None) -> str:
        return super().to_eng_string(context=context)

    def __format__(
        self, specifier: str, context: Context | None = None
    ) -> str:
        return super().__format__(specifier, context)

    # * comparison

    def __eq__(self, value: object) -> bool:  # type: ignore
        if not isinstance(value, Numeric):
            return False
        return super().__eq__(decimal.from_numeric(value))

    def __ne__(self, value: object) -> bool:  # type: ignore
        if not isinstance(value, Numeric):
            return True
        return super().__ne__(decimal.from_numeric(value))

    def __ge__(self, value: Numeric) -> bool:  # type: ignore
        return super().__ge__(decimal.from_numeric(value))

    def __gt__(self, value: Numeric) -> bool:  # type: ignore
        return super().__gt__(decimal.from_numeric(value))

    def __le__(self, value: Numeric) -> bool:  # type: ignore
        return super().__le__(decimal.from_numeric(value))

    def __lt__(self, value: Numeric) -> bool:  # type: ignore
        return super().__lt__(decimal.from_numeric(value))

    # * sign

    def __abs__(self) -> decimal:
        return decimal(super().__abs__())

    def __neg__(self) -> decimal:
        return decimal(super().__neg__())

    def __pos__(self) -> decimal:
        return decimal(super().__pos__())

    # * addition

    def __add__(self, value: Numeric) -> decimal:
        return decimal(super().__add__(decimal.from_numeric(value)))

    def __radd__(self, value: Numeric) -> decimal:
        return decimal(super().__radd__(decimal.from_numeric(value)))

    # * subtraction

    def __sub__(self, value: Numeric) -> decimal:
        return decimal(super().__sub__(decimal.from_numeric(value)))

    def __rsub__(self, value: Numeric) -> decimal:
        return decimal(super().__rsub__(decimal.from_numeric(value)))

    # * multiplication

    def __mul__(self, value: Numeric) -> decimal:
        return decimal(super().__mul__(decimal.from_numeric(value)))

    def __rmul__(self, value: Numeric) -> decimal:
        return decimal(super().__rmul__(decimal.from_numeric(value)))

    # * division

    def __truediv__(self, value: Numeric) -> decimal:
        return decimal(super().__truediv__(decimal.from_numeric(value)))

    def __rtruediv__(self, value: Numeric) -> decimal:
        return decimal(super().__rtruediv__(decimal.from_numeric(value)))

    def __floordiv__(self, value: Numeric) -> decimal:
        return decimal(super().__floordiv__(decimal.from_numeric(value)))

    def __rfloordiv__(self, value: Numeric) -> decimal:
        return decimal(super().__rfloordiv__(decimal.from_numeric(value)))

    # * modulo

    def __mod__(self, value: Numeric) -> decimal:
        return decimal(super().__mod__(decimal.from_numeric(value)))

    def __rmod__(self, value: Numeric) -> decimal:
        return decimal(super().__rmod__(decimal.from_numeric(value)))

    def __divmod__(self, value: Numeric) -> tuple[decimal, decimal]:
        dt = super().__divmod__(decimal.from_numeric(value))
        return (decimal(dt[0]), decimal(dt[1]))

    def __rdivmod__(self, value: Numeric) -> tuple[Decimal, Decimal]:
        raise NotImplementedError()

    # * exponentation / logarithms

    def __pow__(self, value: Numeric, mod: Numeric | None = None) -> decimal:
        mod = decimal(mod).decimal if mod else None
        return decimal(
            super().__pow__(decimal.from_numeric(value).decimal, mod)
        )

    def __rpow__(self, value: Numeric, mod: Context | None = None) -> decimal:
        return decimal(
            super().__rpow__(decimal.from_numeric(value).decimal, mod)
        )

    def exp(self, context: Context | None = None) -> decimal:
        return decimal(super().exp(context))

    def sqrt(self, context: Context | None = None) -> decimal:
        return decimal(super().sqrt(context))

    def ln(self, context: Context | None = None) -> decimal:
        return decimal(super().ln(context))

    def log10(self, context: Context | None = None) -> decimal:
        return decimal(super().log10(context))

    def logb(self, context: Context | None = None) -> decimal:
        return decimal(super().logb(context))

    # * rounding

    def __round__(self, ndigits: int = 0) -> decimal:  # type: ignore
        return decimal(super().__round__(ndigits))

    def __floor__(self) -> int:
        return super().__floor__()

    def __ceil__(self) -> int:
        return super().__ceil__()

    # * stats
    def max(self, other: Numeric, context: Context | None = None) -> decimal:
        return decimal(super().max(decimal(other).decimal, context))

    def min(self, other: Numeric, context: Context | None = None) -> decimal:
        return decimal(super().min(decimal(other).decimal, context))

    # TODO
    @property
    def real(self) -> Self: ...
    @property
    def imag(self) -> decimal: ...
    def __trunc__(self) -> int: ...
    def conjugate(self) -> decimal: ...
    def remainder_near(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def normalize(self, context: Context | None = None) -> decimal: ...
    def fma(
        self, other: Numeric, third: Numeric, context: Context | None = None
    ) -> decimal: ...
    def quantize(
        self,
        exp: Numeric,
        rounding: str | None = None,
        context: Context | None = None,
    ) -> decimal: ...
    def same_quantum(
        self, other: Numeric, context: Context | None = None
    ) -> bool: ...
    def to_integral_exact(
        self, rounding: str | None = None, context: Context | None = None
    ) -> decimal: ...
    def to_integral_value(
        self, rounding: str | None = None, context: Context | None = None
    ) -> decimal: ...
    def to_integral(
        self, rounding: str | None = None, context: Context | None = None
    ) -> decimal: ...
    def adjusted(self) -> int: ...
    def canonical(self) -> decimal: ...
    def compare_signal(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def compare_total(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def compare_total_mag(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def copy_abs(self) -> decimal: ...
    def copy_negate(self) -> decimal: ...
    def copy_sign(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def is_canonical(self) -> bool: ...
    def is_finite(self) -> bool: ...
    def is_infinite(self) -> bool: ...
    def is_nan(self) -> bool: ...
    def is_normal(self, context: Context | None = None) -> bool: ...
    def is_qnan(self) -> bool: ...
    def is_signed(self) -> bool: ...
    def is_snan(self) -> bool: ...
    def is_subnormal(self, context: Context | None = None) -> bool: ...
    def is_zero(self) -> bool: ...
    def logical_and(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def logical_invert(self, context: Context | None = None) -> decimal: ...
    def logical_or(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def logical_xor(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def max_mag(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def min_mag(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def next_minus(self, context: Context | None = None) -> decimal: ...
    def next_plus(self, context: Context | None = None) -> decimal: ...
    def next_toward(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def number_class(self, context: Context | None = None) -> str: ...
    def radix(self) -> decimal: ...
    def rotate(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def scaleb(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def shift(
        self, other: Numeric, context: Context | None = None
    ) -> decimal: ...
    def __reduce__(self) -> tuple[type[Self], tuple[str]]: ...
    def as_tuple(self) -> DecimalTuple: ...


Numeric = int | float | Decimal | decimal  # | Fraction | Rational
Realish = Numeric | str

N = TypeVar("N", bound=Numeric)

from typing import Generic, TypeVar, Self, Literal
from enum import Enum

U = TypeVar("U", bound=LiteralString)

# U = TypeVar("U")

# # TODO: count of type T with prefix P - Generic[T, P]
# class count(int, Generic[U]):
#     __unit: U
#     __type: type[U]

#     @property
#     def unit(self) -> U:
#         return self.__unit

#     @property
#     def type(self) -> type[U]:
#         if self.__type is str:
#             return Literal[self.__unit] # type: ignore
#         raise
#         # return self.__type

#     def __new__(cls, value: int, unit: U, t: type[U] | None = None) -> count[U]:
#         inst = super(count, cls).__new__(cls, value)
#         inst.__unit = unit
#         inst.__type = t if t else type(unit)
#         return inst

#     def __repr__(self) -> str:
#         # return f"{super().__repr__()} [{self.__type.__name__}:{self.__unit}]"
#         return f"{super().__repr__()} [{self.__unit}]"

#     def __eq__(self, other) -> bool: ...

# U = TypeVar("U", bound=LiteralString)

# class _count(int, Generic[U]):
#     __unit: U

#     def __repr__(self) -> str:
#         return f"{super().__repr__()} [{self.__unit}]"

#     def __iter__(self):
#         for k, v in dict(value=int(self), unit=self.unit).items():
#             yield k, v

#     def __new__(cls, value: int, unit: U) -> count[U]:
#         inst = super(count, cls).__new__(cls, value)
#         inst.__unit = unit
#         return inst

#     @property
#     def unit(self) -> U:
#         return self.__unit

#     @unit.setter
#     def unit(self, u: U) -> None:
#         ...


class even(int): ...


class odd(int): ...


class natural(int): ...


# ? positive, pos
class positive(Generic[N]): ...


# ? negative, neg
class negative(Generic[N]): ...


class ratio(pair[int]): ...


# 2d point default?
class point2(pair[decimal]): ...


# TODO: should be float pair option iff importance(speed) > importance(accuracy)
# class point(pair[real]):
#     """
#     n-dim point
#     """
#     ...


class fpoint(pair[float]): ...


class ndpoint(tuple[decimal, ...]):
    @property
    def dimensions(self) -> positive[int]: ...



"""
TODO: doc
"""

# * lib types
from ._count import (
    count,
    ConvertibleToCount,
    # BitCount,
    # ByteCount,
)
from ._source import (
    definition,
    ConvertibleToDefinition,
    module,
    ConvertibleToModule,
)
from ._extended import estr
from ._path import (
    ConvertibleToPath,
    FilePath,
    DirectoryPath,
    Path,
    PathLike,
)
from ._model import Model, Documented, Blob

# * other common types from various standard libs/dependencies

# NOTE: typing_extensions is used over the default `typing` module for
# more compatibility with pydantic for these types
from typing_extensions import (
    TypedDict,  # TODO: add others where pydantic requires typing_extensions type
)

from typing import (
    Any,
    NoReturn,
    Never,
    Self,
    LiteralString,
    ClassVar,
    Final,
    TypeVar,
    Union,
    Optional,
    Literal,
    TypeAlias,
    Concatenate,
    TypeGuard,
    TypeIs,
    ForwardRef,
    overload,
    # TODO: add others as needed
)


class of:
    """
    Various type variables. Called `of` so can be used like:

    ```python
    import typically as t

    # t.of.M - as in: "type of M"
    def load(path: t.ConvertibleToPath, model: type[t.of.M]) -> t.of.M:
        ...
    ```
    """

    M = TypeVar("M", bound=Model)



# TODO | impl: https://en.wikipedia.org/wiki/List_of_types_of_numbers

from __future__ import annotations
import builtins
import typing
from typing import Generic, TypeVar, Union
from typing_extensions import Self
from ._guard import require
from ._typeshed import ConvertibleToInt, ConvertibleToFloat


class whole(int):
    def __new__(cls, x: ConvertibleToWhole, /) -> Self:
        require.ge(x, 0, "Whole numbers must be greater than or equal to 0.")
        return int.__new__(cls, x)


ConvertibleToWhole = ConvertibleToInt | whole

numeric = Union[
    int,
    float,
    # real,
    # +numpy types
    # +decimal, fraction, etc
]
NT = TypeVar("NT", bound=numeric)


class positive(Generic[NT]):
    # __value: N
    # __type: type[N]
    # def __new__(cls, value: N) -> pos[N]:
    #     return type(value).__new__(type(value), value)
    def __init__(self, value: NT) -> None:
        require.gt(value, 0)
        self.__value = value
        self.__type = type(value)
        raise NotImplementedError("TODO")

    @property
    def type(self) -> type[NT]:
        return self.__type

    @property
    def value(self) -> NT:
        return self.__value


class negative(Generic[NT]): ...


class ordinal: ...


# natural starts at 0, or 1, dependending on what branch of math one is operating in it seems
# ref: https://en.wikipedia.org/wiki/Natural_number
# class natural(int): ...


class cardinal: ...


class nominal: ...



import builtins
from typing import TypeVar, Any

T = TypeVar("T")
V = TypeVar("V")
KT = TypeVar("KT")

collection = set[T] | list[T] | tuple[T, ...]
ordered = list[T] | tuple[T, ...]


def coerce(f): ...


class utils:
    """local utility functions"""

    # TODO: replace/move
    @staticmethod
    def emsg(msg: str, detail: str | None) -> str:
        """concat msg + detail"""
        return msg if msg is None else msg + f". {detail}"


class check:  # ? compare, condition
    """Returns bools"""

    @staticmethod
    def haskeys(value: dict[KT, Any], keys: set[KT]) -> bool: ...


class verify:  # ? verify, validate
    """Returns validated type"""

    @staticmethod
    def isinstance(value: V, expected: type[V]) -> V:
        """

        Args:
            value (object): _description_
            expected (type): _description_

        Raises:
            TypeError: _description_
        """
        # NOTE: clip value to avoid console overflow
        value_str_clip = str(value)
        value_str_clip: str = (
            value_str_clip
            if len(value_str_clip) < 128
            else f"{value_str_clip[:128]}..."
        )
        value_cls_name: str = value.__class__.__qualname__
        exp_cls_name: str = expected.__qualname__
        if not builtins.isinstance(value, expected):
            raise TypeError(
                f"Expected '{exp_cls_name}' but received '{value_cls_name}' with value '{value_str_clip}'"
            )
        return value

    @staticmethod
    def issubclass():
        raise NotImplementedError()


# TODO: have o1 go through and add other extra utils


class IsEqualError(ValueError):
    def __init__(
        self,
        source: Any,
        target: Any,
        detail: str | None = None,
        *args: builtins.object,
    ) -> None:
        self.source = source
        self.target = target
        self.detail = detail
        # msg = utils.emsg(f"'{source}' cannot be equal to '{target}'", detail)
        msg = utils.emsg(f"'{source}' is equal to '{target}'", detail)
        super().__init__(msg, *args)


class NotEqualError(ValueError):
    def __init__(
        self,
        source: Any,
        target: Any,
        detail: str | None = None,
        *args: builtins.object,
    ) -> None:
        self.source = source
        self.target = target
        self.detail = detail
        msg = utils.emsg(f"'{source}' does not equal '{target}'", detail)
        super().__init__(msg, *args)


class NotGreaterThanOrEqualToError(ValueError):
    def __init__(
        self,
        source: Any,
        target: Any,
        detail: str | None = None,
        *args: builtins.object,
    ) -> None:
        self.source = source
        self.target = target
        self.detail = detail
        msg = utils.emsg(
            f"'{source}' is not greater than or equal to '{target}'", detail
        )
        super().__init__(msg, *args)


class NotGreaterThanError(ValueError):
    def __init__(
        self,
        source: Any,
        target: Any,
        detail: str | None = None,
        *args: builtins.object,
    ) -> None:
        self.source = source
        self.target = target
        self.detail = detail
        msg = utils.emsg(f"'{source}' is not greater than '{target}'", detail)
        super().__init__(msg, *args)


class param:
    """
    Guarding parameters/attributes.
    """

    ...


class require:
    """Returns none, raises if false"""

    @staticmethod
    def eq(
        source: Any,
        target: Any,
        detail: str | None = None,
        raises: type[NotEqualError] = NotEqualError,
        *args,
        **kwargs,
    ) -> None:
        # if not (source == target):
        # if source != target:
        #     raise raises(source, target, detail, *args, **kwargs)
        if source == target:
            return
        raise raises(source, target, detail, *args, **kwargs)

    @staticmethod
    def ne(
        source: Any,
        target: Any,
        detail: str | None = None,
        raises: type[IsEqualError] = IsEqualError,
        *args,
        **kwargs,
    ) -> None:
        # if not (source != target):
        # if source == target:
        #     raise raises(source, target, detail, *args, **kwargs)
        if source != target:
            return
        raise raises(source, target, detail, *args, **kwargs)

    @staticmethod
    def ge(
        source: Any,
        target: Any,
        detail: str | None = None,
        raises: type[
            NotGreaterThanOrEqualToError
        ] = NotGreaterThanOrEqualToError,
        *args,
        **kwargs,
    ) -> None:
        if source >= target:
            return
        raise raises(source, target, detail, *args, **kwargs)

    @staticmethod
    def gt(
        source: Any,
        target: Any,
        detail: str | None = None,
        raises: type[NotGreaterThanError] = NotGreaterThanError,
        *args,
        **kwargs,
    ) -> None:
        if source > target:
            return
        raise raises(source, target, detail, *args, **kwargs)

    @staticmethod
    def lt(): ...
    @staticmethod
    def le(): ...

    @staticmethod
    def notinstance(
        source: object,
        target: type,
        detail: str | None = None,
    ) -> None:
        # stype: str = source.__class__.__qualname__
        ttype: str = target.__qualname__
        if builtins.isinstance(source, target):
            raise TypeError(f"'{source}' cannot be an instance of '{ttype}'")

    # TODO: make sure this can handle generics, unions, literal, etc
    @staticmethod
    def isinstance(
        source: object,
        target: type,
        detail: str | None = None,
        raises: type[ValueError] = ValueError,
    ) -> None:
        ttype: str = target.__qualname__
        if not builtins.isinstance(source, target):
            # raise TypeError(
            #     f"Expected '{ttype}' but received '{stype}' with value '{value_str_clip}'"
            # )
            raise TypeError(f"'{source}' must be an instance of '{ttype}'")

    # @staticmethod
    # def contains(
    #     domain: collection[T], member: T
    # ) -> None:  # ? co & nc for negated? - would be consistent
    #     """
    #     Whether the domain of items contains the item in question.
    #     """
    #     # future: for single item, diff from 'has' which maybe should do set intersection?
    #     # future: expand to sequences? would need to figure out order, eg: contains("hello world", "hello") vs contains("hello world", "hw")
    #     if member not in domain:
    #         msg = f"'{member}' must be a member of '{domain}'"
    #         raise ValueError()

    # @staticmethod
    # def has():
    #     """same as 'hasall'"""
    #     ...

    # @staticmethod
    # def hasone(): ...

    # @staticmethod
    # def hasnone(): ...  # ? hasnt (hasn't, has not, somewhat medieval)

    # @staticmethod
    # def hassome(): ...

    # @staticmethod
    # def hasall(): ...

    # @staticmethod
    # def hasany() -> None: ...

    # @staticmethod
    # def notin(
    #     value: Any,
    #     forbidden: collection[Any],
    #     detail: str | None = None,
    # ) -> None:
    #     if value in forbidden:
    #         msg = utils.comd(
    #             f"Value '{value}' must not be a member of '{forbidden}'",
    #             detail,
    #         )
    #         raise ValueError(msg)

    # @staticmethod
    # def isin(): ...

    # @staticmethod
    # def istype(): ...

    # @staticmethod
    # def issubclass():
    #     raise NotImplementedError()

    # @staticmethod
    # def haskeys(value: dict[KT, Any], keys: set[KT]) -> None: ...



from __future__ import annotations
import pydantic
from os import PathLike
from pathlib import Path

DirectoryPath = pydantic.DirectoryPath
"""
This is a proxy for `pydantic.DirectoryPath`, which in turn is
just an annotated type for `pathlib.Path`.

It only validates that the path is actually a directory when used
as an attribute in a pydantic model.

Attempting to instantiate it will work but may show in a type error:
`Object of type "Annotated" is not callable`.

It would be used when you need to validate a path is a real directory
path, but only on the instantiation of a pydantic model the path is
an attribute of.
"""

FilePath = pydantic.FilePath
"""
This is a proxy for `pydantic.FilePath`, which in turn is
just an annotated type for `pathlib.Path`.

It only validates that the path is actually a file when used
as an attribute in a pydantic model.

Attempting to instantiate it will work but may show in a type error:
`Object of type "Annotated" is not callable`.

It would be used when you need to validate a path is a real file
path, but only on the instantiation of a pydantic model the path is
an attribute of.
"""

ConvertibleToPath = str | Path
"""
NOTE:
    Not sure what to call this. Would use `PathLike`,
    but already used by `os.PathLike`, and thats an abc,
    so would be confusing.
    
    Ideas:

    - PathLike
    - Pathish
    - ConvertibleToPath

"""


# class RealDirectoryPath(pathlib.Path):
#     def __new__(cls, /, path: Pathish) -> RealDirectoryPath:
#         path = pathlib.Path(path)
#         if not path.is_dir():
#             raise
